#include "X3D_QLoader.h"
#include "QFileElement.h"
#include "X3D_Creator.h"
#include "X3DAbstractNode.h"
#include "X3D_Scene.h"

#include <qxml.h>
#include <qfile.h>
#include <iostream>

using namespace std;

namespace X3DTK {
namespace X3D {

class QSAX2X3DHandler : public QXmlDefaultHandler
{
public: 
  QSAX2X3DHandler(Creator *creator);
  ~QSAX2X3DHandler();
  
  bool startDocument();
  bool startElement(const QString &namespaceURI, const QString &localName, const QString &qName, const QXmlAttributes &atts);
  bool endElement(const QString &namespaceURI, const QString &localName, const QString &qName);
  bool endDocument();
  bool fatalError(const QXmlParseException &exception);
  bool error(const QXmlParseException &exception);
  
  inline X3DAbstractNode *getRoot() const {return _root;};
  inline const ProcessHistory &getHistory() const {return _history;};
  inline const MFRoute &getRoutes() const {return _routes;};
    
private:
  MFNode _nodeStack;
  X3DAbstractNode *_root;
  Creator *creator;
  map<SFString, X3DAbstractNode *> _nameNodeMap;
  unsigned int _depth;
  bool _header;
  ProcessHistory _history;
  MFRoute _routes;
};

QSAX2X3DHandler::QSAX2X3DHandler(Creator *creator)
: _root(0), creator(creator), _depth(0), _header(false)
{
}

QSAX2X3DHandler::~QSAX2X3DHandler()
{
}

bool QSAX2X3DHandler::startDocument()
{
  _root = 0;
  _depth = 0;
  _header = false;
  _nodeStack.push_back(0);
  _history.getProcessMap().clear();
  _routes.clear();
  
  return true;
}

bool QSAX2X3DHandler::startElement(const QString &, const QString &, const QString &qName, const QXmlAttributes &atts)
{
  //_depth == 0, meaning that all the fathers of the node are defined nodes(have a Creation Function)
  //This variable is used to jump a sub-tree of the graph scene, when an unknown node is met.
  
  if (SFString((const char *)qName) == SFString("X3D"))
    return true;

  //////////////////////////// header ////////////////////////////////////////////////

  if (SFString((const char *)qName) == SFString("head"))
  {
    _header = true;
    return true;
  }  
  
  if (_header)
  {
    if (SFString((const char *)qName) == SFString("meta"))
    {
      QFileElement e((const char *)qName, &atts);
      int index, indexDate;  
      index = e.getIndexAttribute("content");
      indexDate = e.getIndexAttribute("date");
      if (index != -1)
      {
        if (indexDate != -1)
          _history.addEntry(e.getAttribute(index), ProcessHistory::loadTime(e.getAttribute(indexDate)));
      }
    }
    return true;  
  }  
  //////////////////////////// Route /////////////////////////////////////////////////
  
  if (SFString((const char *)qName) == SFString("ROUTE"))
  {
    int index;
    X3DAbstractNode *fromNode = 0;
    X3DAbstractNode *toNode = 0;
    
    SFString fromField, toField;
    QFileElement e((const char *)qName, &atts);
    
    index = e.getIndexAttribute("fromNode");
    if (index != -1)
    {
      map<SFString, X3DAbstractNode *>::const_iterator it = _nameNodeMap.find(e.getAttribute(index));
      if (it != _nameNodeMap.end())
        fromNode = (*it).second;
      else
        fromNode = 0;
    }
    
    index = e.getIndexAttribute("toNode");
    if (index != -1)
    {
      map<SFString, X3DAbstractNode *>::const_iterator it = _nameNodeMap.find(e.getAttribute(index));
      if (it != _nameNodeMap.end())
        toNode = (*it).second;
      else
        toNode = 0;
    }
    
    index = e.getIndexAttribute("fromField");
    if (index != -1)
      fromField = e.getAttribute(index);
      
    index = e.getIndexAttribute("toField");
    if (index != -1)
      toField = e.getAttribute(index); 
    
    X3DRoute *r = X3DRoute::create(fromNode, fromField, toNode, toField);  
    if (r != 0)
      _routes.push_back(r);   
    
    return true;  
  }
  
  //////////////////////////// Scene /////////////////////////////////////////////////

  if (_depth == 0)
  {
    QFileElement e((const char *)qName, &atts);

    int index;  
      SFNode Father = _nodeStack.back();
      SFNode N;
      index = e.getIndexAttribute("USE");
      
      if (index != -1)
      {
        map<SFString, X3DAbstractNode *>::const_iterator it = _nameNodeMap.find(e.getAttribute(index));
        if (it != _nameNodeMap.end())
          N = (*it).second;
        else
          N = 0;  
      }  
      else  
      {
        X3D::X3DNode *cN = creator->createFromName(e.getName());
        if (cN == 0)
        {
          _depth = 1;
          return true;
        }
        cN->loadAttributes(&e);
        N = cN;
      }
      index = e.getIndexAttribute("DEF");
      if (index != -1)
      {
        N->setName(e.getAttribute(index));
        _nameNodeMap[e.getAttribute(index)] = N;
      }  

      if (Father != 0)
        Father->setChild(N);
      else
        _root = N;  
      _nodeStack.push_back(N);  
  }
  else
    ++_depth;
    
  return true;  
}

bool QSAX2X3DHandler::endElement(const QString &, const QString &, const QString &qName)
{
  if (SFString((const char *)qName) == SFString("head"))
  {
    _header = false;
    return true;
  }
  
  if (_header)
    return true;
  
  if ((_depth == 0) && (_nodeStack.size() > 0))
    _nodeStack.pop_back();
  else
    --_depth;  
  
  return true;
}

bool QSAX2X3DHandler::endDocument()
{
  _nodeStack.clear();
  _nameNodeMap.clear();

  return true;
}


bool QSAX2X3DHandler::error(const QXmlParseException &exception)
{
  cx3d << "QLoader::error: " << exception.message() << " at line: " << exception.lineNumber() << endl;
  return true;     
}

bool QSAX2X3DHandler::fatalError(const QXmlParseException &exception)
{
  cx3d << "QLoader::fatalError: " << exception.message() << " at line: " << exception.lineNumber() << endl;
  return false;     
}


class QLoaderImplementation
{     
public:
  QLoaderImplementation(){};
  
  QXmlSimpleReader *_parser;
  QSAX2X3DHandler *_handler;
};

QLoader::QLoader(Creator *creator)
: X3DXmlLoader(creator), _impl(new QLoaderImplementation())
{
  _impl->_parser = new QXmlSimpleReader();
  _impl->_handler = new QSAX2X3DHandler(creator);
  _impl->_parser->setContentHandler(_impl->_handler);
  _impl->_parser->setErrorHandler(_impl->_handler);
}

QLoader::~QLoader()
{
  delete _impl->_parser;
  delete _impl->_handler;
}

X3D::Scene *QLoader::load(const char *file, bool) const 
{
  SFString sfile(file);
  
  // Goes to the directory of the file.
  char *cwd = (char *)malloc(65535);
  char *value = getcwd(cwd, 65535);
  if (value == 0)
  {
    cx3d << "QLoader::load: unable to get current working directory!" << endl;
    return 0;
  }  
  
  if (chdir(sfile.getPath()) != 0)
  {
    cx3d << "QLoader::load: unable to set the current working directory to !" << sfile.getPath()<< endl;
    return 0;
  }
  
  QFile xmlFile((const char *)sfile.getFile());
  QXmlInputSource source(&xmlFile);   
  
  try 
  {
    _impl->_parser->parse(source);
  }
  catch (...) 
  {
    cout << "QLoader::load: unexpected Exception!" << endl;
  }
  
  // Goes back to the initial directory.
  if (chdir(cwd) != 0)
  {
    cx3d << "QLoader::load: unable to set the current working directory to !" << cwd << endl;
    return 0;
  }
 
  free(cwd);
  
  X3D::Scene *scene = dynamic_cast<X3D::Scene *>(_impl->_handler->getRoot());
  if (scene != 0)
  {
    scene->setFileName(file);
    ProcessHistory::loadScene(scene, _impl->_handler->getHistory());
    X3DRoute::loadScene(scene, _impl->_handler->getRoutes());
  } 
  
  return scene;
}

}
}
